# LDAP\_INP

## Introduction

Le principe de ce module est de fournir une sorte d'ORM fait maison qui permet de convertir des objets Python en objets LDAP et réciproquement.

## LDAP, késako ?

Le LDAP est un protocole réseau qui permet d'accéder/stocker des données hiérarchiques. Ce protocole est particulièrement adapté aux données qui sont accédées souvent en lecture mais rarement en écriture, contrairement à une base de donnée SQL standard. Un des intérêts du LDAP est que les données sont hiérarchisées tout en suivant un schéma qui permet de spécifier le type et les attributs chaque objet. Cela permet de s'assurer une certaine cohérence et se prête bien à des problèmatiques de gestion d'utilisateurs ou de groupes. Pour plus d'informations, je vous invite à vous reporter à la [présentation](https://www.bde.enseeiht.fr/clubs/net7/supportFormations/ldap/ldap.pdf) de 2011 de net7 à ce sujet.

Le schéma d'INP-net est consultable sur le [projet conf_ldap](https://git.inpt.fr/inp-net/conf_ldap/). Dans la version courante à l'heure où j'écris ces lignes (septembre 2019), un élève est décrit de la manière suivante:

```
objectclass ( 1.3.6.1.4.1.15323.1.1
	NAME 'Eleve'
	DESC 'Représente un élève scolarisé dans une école de l INP. Associé à un compte INP-net (posixAccount) et à une adresse mail (from qmailUser). Les lien n est pas mis car ldap ne permet pas de faire qu un objet de type STRUCTURAL hérite d un type AUXILIARY)).'
	SUP inetOrgPerson
	STRUCTURAL
	MUST ( promo $ filiere $ ecole $ genre $ inscritAE $ loginShell $ userPassword )
	MAY ( aim $ icq $ googleTalk $ jabber $ msn $ skype $ yahoo $ personalWebsite $ parentsAddress $ mailEcole $ mailAnnexe $ birthdate $ loginTP $ hasWebsite $ nickname $ givenNameSearch $ snSearch $ uidParrain $ subscribedate $ mailLocal $ inscritFrappe $ sshKey $ inscritPassVieEtudiant $ finPassVieEtudiant )
)
```

``objectClass`` signale que nous déclarons un nouveau type d'objet.
Ce type d'objet dispose d'un identifiant unique (``1.3.6.1.4.1.15323.1.1``), et ces identifiants sont eux-mêmes organisés hiérarchiquement.

Un Élève est donc un objet qui contient des champs obligatoires (via le mot clé ``MUST``, qui oblige les champs associés à être non-nuls), et des champs optionnels (``MAY``).
La ligne ``SUP inetOrgPerson`` indique qu'un Élève descend d'un objet ``inetOrgPerson`` et hérite donc de ses propriétés.

Comme indiqué dans la description de l'objectClass Eleve, un élève tel que stocké actuellement dans notre LDAP fait partie d'autres objectClass comme ``qmailUser`` ou ``PosixAccount`` qui servent respectivement à pouvoir régler l'adresse mail (en théorie, on peut aussi stocker certains paramètres de mail mais net7 n'en fait pas usage). Finalement, un objet LdapEleve (défini dans ``__init__.py``) fait donc partie des types d'objets suivant: ``["top", "person", "organizationalPerson", "inetOrgPerson", "qmailUser", "posixAccount", "shadowAccount", "Eleve"]``.

## Fonctionnement de LDAP\_INP

### Types d'objets

Les différents types d'objets qui existent actuellement (et leurs relations d'héritage) sont:

* LdapObject
    * LdapEcole
    * LdapFiliere
    * LdapAlias
    * LdapCompte
    * LdapEleveNomPrenom
        * LdapEleve
            * LdapEleveAvecPhoto
            * LdapEleveAvecPwd
                * LdapEleveAvecPhotoEtPwd
    * LdapPosixGroup
        * LdapGroupe
            * LdapClub
    * LdapCompteCri

Tous les objets LDAP sont donc des enfants de ``LdapObject``, qui est une classe implantée dans ``ldapobject.py``.

## LdapObject

``LdapObject`` sert donc, comme on viens de le voir, à abstraire des traits communs à plusieurs objets pour pouvoir les manipuler de manière cohérentes et à structurer ainsi les données. Pour ce faire, les classes héritant de ``LdapObject`` possèdent un champ ``objects`` qui permettent de maniupler les objets de ce type. Par exemple, ``LdapEleve.objects.filter(ecole=ecole_ensat)`` est une object qui contient tous les élèves qui sont à l'ENSAT.

Prenons un exemple:

```
class LdapFoo(ldapobject.LdapObject):
    directory = "principal"
    manager = "root"
    objectClass = ["top", "organization", "Foo"]

    fooAttr = ldapobject.LdapStringField()
```

Notons que ``directory`` et ``manager`` servent à spécifier le compte et le serveur LDAP à utiliser pour ce type d'objet. Les logins sont récupérés à partir de ces deux identifiants dans conn.py: ``(uid, password) = LDAP_ACCESS_SECRET[directory]['accounts'][user]``.

D'un autre côté ``objectClass`` spécifie les classes d'objet LDAP auxquelles notre objet Python appartient. Cela est doublement utile:

- lors de la création d'objets LDAP depuis python car nous avons la garantie que l'objet créé possède les propriétés nécessaires (grâce aux fameux ``MUST`` dont on a parlé plus haut), sinon le serveur LDAP refuse la création de l'objet
- lors de la recherche d'objets LDAP, seul les objets du bon type seront ainsi recherchés

``LdapObject`` convertit les données dans un format lisible pour le serveur LDAP d'un côté et en objets python de l'autre. Ces conversions ont lieu dans les méthodes ``__init_from_ldap`` (LDAP -> python) et ``__init_from_user`` (python -> LDAP). Pour les convertir, elle se sert des méhodes ``python_to_ldap``, ``ldap_to_python`` et ``cast`` définies dans les classes héritant de ``LdapBaseField``.

### Création d'une classe héritée de LdapObject

La classe ``LdapObject``  possède une Métaclasse qui permet de traiter les champs des objets qui héritent de ``LdapObject``. Lorsque l'interpréteur python lit la description d'une classe qui hérite de ``LdapObject``, il appelle le constructeur ``__new__`` de ``LdapObjectMetaclass``, qui va initialiser et valider les champs et les attributs de la classe qui nous intéresse. 

Cette classe est configurée de la manière suivante:

1. La fonction ``LdapObjectMetaclass.__new__`` est appelée à la déclaration de le classe (et non pas de chaque objet de la classe: une métaclasse est à une classe ce qu'une classe est à un objet)
2. On récupère les attributs de la classe LdapFoo ou de ses classes parentes (dans notre cas particulier: ``directory``, ``manager``, ``objectClass`` et ``fooAttr``) dans un dictionnaire ``attributes``.
Plus précisément, on itére sur chaque attribut, et pour chaque champ LDAP (tout champ qui hérite de ``LdapBaseField``), on l'ajoute à ce dictionnaire ``attributes``. Si ce n'est pas un champ LDAP, on l'ajoute au dictionnaire des propriétés de notre classe (``LdapFoo``).
``objectClass`` jouit d'un traitement un peu particulier pour s'assurer de récupérer la version la plus spécifique possible ou une valeur par défaut héritée d'une classe parente si on n'en trouve aucune dans la classe courante.
Les champs de type ``LdapRDNField`` servent quant à eux à identifier de façon unique un champ et on s'assure qu'il n'y en a qu'un par classe.
3. On effectue quelques vérifications pour garantir que toutes les classes qui héritent de ``LdapObject `` (à l'exception de ``LdapObject`` qui est vide) sont correctement complétées dans ``__init__.py``. En particulier, on vérifie que :
    * L'objet possède bien des objectClass
    * L'objet possède un champ d'identification (le fameux ``LdapRDNFiled`` dont nous parlions précédemment)
    * L'objet est associé à un serveur ldap (<=> il a un attribut ``directory``)
    * L'identifiant complet de l'objet n'est pas spécifié dans la classe, pour éviter de sérieux problèmes ultérieurement
4. Toujours pour ces classes qui héritent de ``LdapObject``, on génère un filtre de base pour obtenir tous les objets qui ont les mêmes ``objectClass`` que la classe que l'on traite.
La manière dont on le génère est la suivante: on prends le premier ``objectClass`` dans la liste ``object_class``, puis on ajoute l'ensemble des autres ``objectClass`` de la liste comme contraintes de recherche au filtre (on utilise la classe ``L`` que l'on détaillera plus bas).
Dans le cas de ``LdapFoo``, on obtient le filtre correspondant à la contrainte LDAP ``(&(objectClass=top)(objectClass=organization)(objectClass=Foo))``.
5. Pour finir, on assigne tous les attributs restants à la classe, on créée la classe, on instance un object ``LdapObjectManager`` - que l'on va également détailler par la suite - et on retourne la classe ainsi configurée.

### LdapObjectManager

``LdapObjectManager`` est une classe qui sert à abstraire les différents types de requètes envoyées à la librairie LDAP. Chaque classe dérivée de ``LdapObject`` possède ainsi un champ ``objects``:

``klass.objects = LdapObjectManager(klass, base_filter)``

De cette manière, on peut ensuite effectuer des actions sur tous les objets de cette classe, comme on l'a vu dans la section sur ``LdapObject``. La liste des fonctions disponible à l'heure où j'écris ces lignes est:

- ``all``: retourne tous les objets
    - exemple: ``LdapEleve.objects.all()``
- ``filter``: retourne tous les objets qui répondent à un ou des prédicats
    - exemple: ``LdapEleve.objects.filter(ecole=ecole_ensat, promo=2022)``
- ``get``: retourne un unique objet qui réponds à un ou des prédicats. Attention, un seul objet doit répondre à ces conditions, sinon il faut utiliser `filter``.
    - exemple: ``Club.objects.get(ident=club_ident, ecole__ident=club_ecole)\\
- ``getDN``: retourne un objet à partir de sa localisation dans l'arbre LDAP (son ``DN``)
    - exemple: ``LdapEleveAvecPwd.objects.getDN(eleve.dn)``
- ``_save``: Crée ou enregistre un objet, suivant s'il existe déjà ou non
    - exemple: ``logement.save()`` (la fonction ``save`` de ``LdapObject`` appele ``_save``)
- ``_delete``: Supprime un objet identifié par son ``DN``
    - exemple: ``logement.delete()`` (la fonction ``delete`` de ``LdapObject`` appele ``_delete``)


``LdapObjectManager`` possède (ou plus précisément partage) un champ ``_connection_pool`` qui sert à mettre en cache les connections ouvertes vers le serveur LDAP, afin de diminuer le nombre de connexions ouvertes et d'améliorer les performances (principalement diminuer la charge liée à l'ouverture de connexions et réduire la latence). Pourquoi l'emploi du mot "partager" ? Parce que ce fonds de connexions a une granularité toute particulière: chaque thread partage le même fonds. De cette manière la réutilisation des connexions est maximale sans introduire des problèmes de parallèlisme.

La gestion de ces connexions est «automatique»: toutes les fonctions - fournies par ``LdapObjectManager`` - décrites ci-dessus s'occupent de se connecter au serveur LDAP si la connexion n'a pas encore été établie ou d'en rouvrir une si le temps d'exécution de la requète dépasse un seuil et que la connexion est alors considérée comme interrompue.

En effet, toutes ces fonctions appellent ``self.try_conn_then_op``, qui est en charge de récupérer une connexion (quitte à en établir une nouvelle), puis d'effectuer l'opération qui lui est passée en paramètre. ``try_conn_then_op`` prend en arguments ``directory``, ``user``, et ``operation``. ``directory`` et ``user`` servent à identifier l'utilisateur LDAP utilisé pour effectuer l'opération (et donc la connexion associée). ``operation`` est une fonction appelée avec en paramètres la connection ainsi ouverte, puis tous les autres paramètres passés à l'appelé - la fonction ``try_conn_then_op`` - reçus via ``*args`` et ``**kwargs``. Cette structure a l'avantage qu'elle permet d'appeler directement une fonction de la librairie LDAP (puisque  ``ldap.ldapobject.LDAPObject.search(conn, ...) <=> conn.search(...)``) mais aussi de définir nos propres fonctions, tel qu'on l'utilise dans la fonction ``_save`` par exemple.


### LdapBaseField et ses dérivés

``LdapBaseField`` est la classe fondamentale dont dérivent les classes ``LdapIntegerField``, ``LdapBooleanField``, ``LdapStringField``, ``LdapRDNField``, ``LdapForeignKey``, ``LdapForeignKeyField``, ``LdapDateField``, ``LdapImageField``. Elle permet de spécifier comment convertir les données entre l'objet LDAP et l'objet python associé (et dans les deux sens, car on veut aussi pouvoir enregistrer des objets pythons dans le LDAP), et ces méthodes de conversions sont appelées dans la classe ``LdapObject``.

### L et LdapObjectQuerySet


#### L

``L`` fournit une abstraction pour la notion filtres LDAP en surchargeant les opérateurs binaires ``__and__`` (``&``), ``__invert__`` (``~``) et ``__or__`` (``|``).

On peut ainsi spécifier un filtre LDAP comme ``(!(&(uid=thobys)(objectClass=LdapEleve)))`` sous la forme ``~(L(uid=thobys)&L(objectClass=LdapEleve))``, et l'appel à la méthode ``__str__`` convertit cet élément en le filtre LDAP correspondant. Ce fait est notamment utilisé dans la méthode ``filter`` de ``LdapObjectManager``.

##### Les modifiers

Un "modifier" est un mot clé postfixé au nom d'un attribut pour indiquer un comportement particulier lors du traitement de l'attribut. Ce mot clé est séparé de l'attribu par deux underscores. Par exemple, ``L(atttribut__modifier1__modifier2=valeur)`` correspond à la recherche de ``attribut=valeur`` en appliquant les modifiers ``modifier1`` et ``modifier2``. Exemple concret: ``L(mailEcole__any__matching=["jean.*@inpt.fr", "jean.dujardin@*.fr"])`` qui sera transcrit en ``(|(mailEcole=jean.*@inpt.fr)(mailEcole=jean.dujardin@*.fr))`` (voir l'effet de chaque modifier plus bas).

Il y a pour le moment trois modifiers: ``matching``, ``all`` et ``any``.

Pour se préserver des attaques par injection LDAP (lors des requètes sur l'annuaire du portail par exemple), les caractères sont échappés par ``L``, sauf si on ajoute le modifier ``__matching`` à un attribut (). Par exemple ``L(objectClass=LdapEleve)&L(uid=thob*)`` sera transcrit en ``(&((objectClass=LdapEleve)(uid=thob)))`` alors que ``L(objectClass=LdapEleve)&L(uid__matching=thob*)`` sera transcrit en ``(&((objectClass=LdapEleve)(uid=thob*)))``.

Deux autres "modifiers" sont disponibles pour rechercher un attribut avec une liste. Ce modifier est même la manière canonique de procéder quand on manipule des attributs dont la valeur est une liste, puisque la sémantique en l'absence de modifier n'est pas évidente: est-ce que ``L(mailEcole=["jean@inpt.fr", "jean.dujardin@bidule.fr"])`` doit correspondre à ``(|(mailEcole=jean@inpt.fr)(mailEcole=jean.dujardin@bidule.fr))`` (l'utilisateur possède l'une de ces addresse) ou à ``(&(mailEcole=jean@inpt.fr)(mailEcole=jean.dujardin@bidule.fr))`` (l'utilisateur possède toutes ces addresse) ? La solution est d'utiliser ``any`` et ``all``:

* ``L(mailEcole__any=["jean@inpt.fr", "jean.dujardin@bidule.fr"])`` devient ``(|(mailEcole=jean@inpt.fr)(mailEcole=jean.dujardin@bidule.fr))`
* ``L(mailEcole__all=["jean@inpt.fr", "jean.dujardin@bidule.fr"])`` devient ``(&(mailEcole=jean@inpt.fr)(mailEcole=jean.dujardin@bidule.fr))`

#### LdapObjectQuerySet

``LdapObjectQuerySet`` est un objet qui représente les réponses à une transaction. Son intérêt consiste à ne pas récupérer systématiquement toutes les réponses à la requête en cours, mais à pouvoir les récupérer à la demande. Pour cela, ``LdapObjectQuerySet`` est un itérateur qui récupère les données par blocs (10 entrées à la fois par défaut) et les cachent pour que l'utilisateur puisse y accéder sans effectuer une requète pour chaque objet, mais sans télécharger toutes les entrées en une seule fois, ce qui est bien souvent inutile et consommateur de bande passante/de CPU.
